(* $Id: gontutil.mli,v 1.13 2001/11/18 23:15:43 malekith Exp $ *)

(* Identifier. *)
type id = string

type location = {mutable loc_file: string; mutable loc_char: int}

(* Types used by gont compiler. *)
type ty =
    Int                         (* integer of the same size as pointer *)
  | Void                        (* unit type *)
  | Bool                        (* boolean, same size as Int *)
  | Float                       (* C's double *)
  | Func of ty * (ty list)
  | String
  | TyVar of id
;;

class type symbol = 
  object
    method get_name : id
    method get_type : ty
    method get_fnc_level : int
    method get_scope_level : int
    method in_closure : bool
    method put_into_closure : unit
  end
  
type op2 =
    IntPlus
  | IntMinus
  | IntMult
  | IntDiv
  | IntMod
  | FloatPlus
  | FloatMinus
  | FloatMult
  | FloatDiv
  | Assign
  | Eq | Ne | Lt | Gt | Ge | Le
  
type op1 =
    IntNegate
  | FloatNegate
  
type parm_record = {parm_name: id;
                    parm_type: ty;
                    mutable parm_symbol: symbol option}

type raw_exp =
    SimpleBinOp of op2 * exp * exp
  | SimpleUnOp of op1 * exp
  | Ref of ref_record   (* reference to variable *)
  | Call of exp * exp list
  | IntConst of string
  | BoolConst of bool
  | FloatConst of string
  | StringConst of string
  | Lambda of lambda_record
  | CrossModuleRef of xmr_record
and exp = {mutable raw_exp: raw_exp; 
           mutable exp_ty: ty option; 
           exp_loc: location}
and ref_record = {ref_name: id;
                  mutable ref_level: int;
                  mutable ref_symbol: symbol option}
and xmr_record = {xmr_module: id;
                  xmr_name: id;
                  mutable xmr_symbol: symbol option}
and raw_stmt =
    Skip
  | Exp of exp
  | Return of exp
  | ReturnVoid
  | VarDef of var_record
  | Block of stmt list
  | Pair of stmt * stmt         (* doesn't introduce new scope *)
  | If of exp * stmt * stmt
  | Loop of stmt
  | BreakIfNot of exp
  | Open of id
and stmt = {raw_stmt: raw_stmt;
            stmt_loc: location}
and var_record = {var_name: id;
                  var_type: ty;
                  mutable var_symbol: symbol option}
and lambda_record = {lambda_ret_type: ty;
                     lambda_parms: parm_record list;
                     lambda_body: stmt}

type fn_decl = {fn_ret_type: ty; 
                fn_name: id; 
                fn_parms: parm_record list;
                fn_body: stmt;
                fn_loc: location;
                mutable fn_symbol: symbol option;
                fn_export_as: string}

type extern_c_decl = {xc_type: ty; 
                      xc_name: id; 
                      xc_loc: location;
                      mutable xc_symbol: symbol option}

type global_decl = {global_type: ty;
                    global_loc: location;
                    global_name: id;
                    mutable global_symbol: symbol option}

type fn_proto_decl = {fp_type: ty;
                      fp_name: id;
                      fp_loc: location;
                      mutable fp_symbol: symbol option}

type decl =
    FuncDecl of fn_decl
  | GlobalDecl of global_decl
  | ExternC of extern_c_decl
  | FuncProto of fn_proto_decl
  | TopLevelOpen of id
;;

exception Syntax_error of location * string
exception Error of string
exception Eof
;;

(* Utils. *)

val reset : unit -> unit

(* Return unique name prefixed with string. String should start
 * with '@', unless it is going to get exported. 
 *)
val seq_name : id -> id

(* Output debugging (trace) message. *)
val debug : string -> unit

(* An alias for Printf.sprintf. *)
val xf : ('a, unit, string) format -> 'a

(* Display warning. *)
val warn : location * string -> unit

(* Display error message. *)
val err_msg : location * string -> unit

(* Raise Syntax_error. *)
val err : location * string -> 'a

(* 'a is type for context, 'b is input value, and 'c is output value. *)
val list_apply : ('a * 'b -> 'a * 'c) -> 'a * 'b list -> 'a * 'c list
;;
